/*
 * led-module - main.c
 *
 * led module control firmware
 *
 * Copyright (C) 2017  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <avr/io.h>
#include <avr/power.h>
#include <avr/wdt.h>
#include <compiler.h>

/*
 * This control program controls the LEDs in the grader led module.
 * It receives enable input signal from higher layer and generates
 * the enable output signal for LED drivers. The led module is designed
 * for pulse operation, continous operation will generate too much heat
 * and overheat LEDs and transistors. This control unit must guarantee
 * that all components are not overheated. We focus only on critical
 * component - MJD44H11 transistor.
 */

/*
 * PERIOD - sampling period, in CPU cycles. The CPU runs at 9.6 MHz.
 * Because we pass-through the enable signal and want very low latency
 * the period must be very low.
 *
 * We use constant sampling period for easier implementation and to
 * avoid interrupt overhead.
 */
#define PERIOD		48	/* 5 us, 200 kHz */

/*
 * MJD44H11 temperature control
 *
 * MJD44H11 thermal model
 *
 *        R P
 * T = ----------
 *      RC s + 1
 *
 * T - temperature over ambient temperature (75 C assumed),
 *     in kelvins (K)
 * C - thermal capacitance, 0.0006 J/K, based on spec
 * R - thermal resistivity, assumed 71.4 K/W, from spec
 * P - transistor power, assumed 10.85 W, maximal power with 25.2 V
 *     input voltage and minimal voltage on LEDs
 *
 * The thermal model has very good accuracy for 200 us to 900 us
 * single pulses, compared to transistor specification. For other
 * pulses predicts higher temperatures.
 *
 * After discretization using backward differences with HZ=200000
 * frequency:
 *
 *         R P
 * T = ----------------
 *      R C (1 - q) HZ + 1
 *
 * where q = z^-1.
 *
 * (R C HZ + 1) T - R C HZ q T = R P
 * (R C HZ + 1) T = R C HZ q T + R P
 * (R C HZ + 1) T = (R C HZ + 1) q T - q T + R P
 *
 *               q T           R P
 * T = q T - ----------- + ------------
 *            R C HZ + 1    R C HZ + 1
 *
 *               q T           R P_MAX E
 * T = q T - ----------- + -----------------
 *            R C HZ + 1       R C HZ + 1
 *
 * To simplify model we want R * P_MAX / (R C HZ + 1) = K = 128
 *
 *                q T'
 * T' = q T' - ------------ + K * E
 *              R C HZ + 1
 *
 * T' = T K (R C HZ + 1) / R P_MAX = T * S
 *
 * R C HZ + 1 = 8569 < 34 * 256
 * (R C HZ + 1) / R P_MAX * K = 1415.33
 *
 * We want 40 C temperature increase, 40 * K = 56633
 */
#define TEMP_HIGH	56633

/*
 * I/O pin mapping
 */
#define PINB_MOSI	(BIT(PB0))
#define PINB_MISO	(BIT(PB1))
#define PINB_SCK	(BIT(PB2))
#define PINB_EN_IN	(BIT(PB3))
#define PINB_EN_OUT	(BIT(PB4))

#ifndef time_after_eq
#define time_after_eq(a, b) \
	((signed char)((signed char)(a) - (signed char)(b)) >= 0)
#endif

void control_loop(void)
{
	unsigned int temp = TEMP_HIGH;
	unsigned char i = 0;
	unsigned char x;
	unsigned char t;

	t = TCNT0;
	for (;;) {
		unsigned char p = 0;

		t += PERIOD;
		do {
			cpu_relax();
		} while (!time_after_eq(TCNT0, t));

		wdt_reset();

		/*
		 * enable LEDs only when it was requested and we are
		 * under the temperature limit
		 */
		x = PINB & PINB_EN_IN;
		if (x && (temp < TEMP_HIGH)) {
			PORTB |= PINB_EN_OUT;
			p = 1;
		} else {
			PORTB &= ~PINB_EN_OUT;
		}

		if (p)
			temp += 128;
		if (++i >= 34) {
			temp -= temp >> 8;
			i = 0;
		}
	}
}

#define TCCR_DIV1		1
static void init(void)
{
	wdt_enable(WDTO_60MS);

	/* just to be sure */
	DDRB = 0;
	PORTB = PINB_MOSI | PINB_SCK;
	DDRB = PINB_EN_OUT | PINB_MISO;

	/* Disable Analog Comparator to save power */
	ACSR = BIT(ACD);

	TCCR0B = TCCR_DIV1;

	clock_prescale_set(clock_div_1);
}

int main(void)
{
	init();
	control_loop();
	return 0;
}
